package jwiki;

import java.io.*;
import java.util.*;
import java.net.*;

public class JHandler implements Runnable {

    private JWiki _parent;
    private Socket _sock;
    // Derrived from the _sock
    private InputStream _in;
    private OutputStream _out;
    // Generic buffer for IO
    private byte _ioBuf[];
    private static final int _maxBuf = 1024*64+1; //Biggest post we can handle.

    private byte _crlf[];


    /** Carriage return **/
    public final static int CR = 0x0D;
    /** Line feed **/
    public final static int LF = 0x0A;

	// Set just after reading the head information.  If the WikiLogin cookie
	// is present then they are logged in.  The login and logindone WikiPostables
	// set the cookie.  Login is good for 6 hours.
	private boolean _userLogin;


    public JHandler(JWiki parent, Socket sock)
    {
        _parent = parent;
		_userLogin = false;
        _sock = sock;
        _ioBuf = new byte[_maxBuf]; // This size is the biggest POST we can take.

        _crlf = new byte[2];
        _crlf[0] = CR;
        _crlf[1] = LF;

        _parent.addHandler(this);
    }

    public InputStream getIn() { return _in; }
    public OutputStream getOut() { return _out; }
    public Socket getSocket() { return _sock; }
    public JWiki getParent() { return _parent; }

    public void run()
    {
        // Bullet proof execution.
        try
        {
            _in = _sock.getInputStream();
            _out = _sock.getOutputStream();
            handleMessage();
        }
        catch(Exception e)
        {
            doDebug(1, "Exception handling a message. (Or killed by JWiki.)");
            System.out.println("exception in JHandler:readLine: " + e.getMessage());
            e.printStackTrace();
            return;
        }
        // Bullet proof closes.
        try { _in.close(); } catch (Exception e1) { }
        try { _out.close(); } catch (Exception e2) { }
        try { _sock.close(); } catch (Exception e3) { }

        _parent.removeHandler(this);
    }

	public boolean allMayDeleteRename()
	{   return JWiki.allMayDeleteRename();
	}

	public void httpHeader() throws IOException
	{   sendString("HTTP/1.0 200 Ok");
        crlf();
	}

    public void writeHeadline(String pageName, boolean unused) throws IOException // Joe Bergin 11/00
	{	String webGif = _parent.webGif();  	
        sendString("<H1>");
        if (webGif != null)
        {	sendString("<IMG SRC=\""+ webGif +"\">");
        }
        sendString("<FONT COLOR=\"BLUE\">"+pageName+"</FONT></H1>\n");
    }

    public void startPage(String title, boolean unused) throws IOException // Joe Bergin 11/00
    {	sendString(JWiki.preambleFirst() + "<TITLE>"+JWiki.titlePrefix()+" "+title+"</TITLE>" + JWiki.preambleRest());    
		writeHeadline(title, unused);
    }

    public void endPage() throws IOException // Joe Bergin 11/00
    {   sendString(JWiki.postambleAll());
    }

    public void minimumEndPage() throws IOException
    {	sendString(JWiki.postambleMinimum());
    }

    public void handleMessage()
        throws Exception
    {
        // Read request first line
        if(_in == null)
        {	doDebug(1, "NULL input in JHandler.handleMessage");
        	return;
        }
        String orig = readLine(_in);
        if(orig.length() == 0)
        {
            doDebug(1, "No HTTP status line (no first line)");
            return;
        }
        doDebug(2, "Request is: " + orig);
        orig = orig.trim();

        int i = orig.indexOf(" ");
        if(i == -1)
        {
            doDebug(1, "No HTTP type");
            return;
        }
        String type = orig.substring(0, i).trim();

        int j = orig.indexOf(" ", i + 1);
        if(j == -1)
        {
            doDebug(1, "No HTTP URL");
            return;
        }
        // Skip the leading / too
        String urls = orig.substring(i + 2, j).trim();

        i = j;
        String vers = orig.substring(i + 1).trim();

        // Read header lines, we only look for Content-Length and the Cookie: WikiLogin
		Hashtable cookies = new Hashtable();
        int contentLength = -1;
        while(true)
        {
            orig = readLine(_in);
            if(orig.length() == 0)
            {
                break;
            }
            orig = orig.trim();
			doDebug(2, "Parsing HTTP header: " + orig);
            String origUpper = orig.toUpperCase();

            i = origUpper.indexOf("CONTENT-LENGTH:");
            if(i == 0)
            {
                i = orig.indexOf(":");
                if(i == -1)
                {
                    doDebug(1, "Invalid HTTP header line, no :");
                    return;
                }

                String val = orig.substring(i + 1).trim();
                try
                {
                    contentLength = Integer.parseInt(val);
                }
                catch(Exception e)
                {
                    doDebug(1, "Invalid Content-Length: " + val);
                    return;
                }
			}

            i = origUpper.indexOf("COOKIE:");
            if(i == 0)
            {
                i = orig.indexOf(":");
                if(i == -1)
                {
                    doDebug(1, "Invalid HTTP header line, no :");
                    return;
                }

                String val = orig.substring(i + 1).trim();
				if(val.indexOf("WikiLogin=true") != -1)
					_userLogin = true;
            }
        }

        // Read any content if there
        if(contentLength > 0)
        {
            doDebug(1, "The request had content length, " + contentLength);
            int n = 0;     // Joe Bergin 11/00
            int remaining = contentLength;
            int total = 0;
            while (remaining > 0)
            {	//System.out.println("Remaining: " + remaining);
        		n = _in.read(_ioBuf, total, remaining);
                remaining -= n;
            	if(n>0)total += n;
                // System.out.println("read: " + n + " total: " + total);
            }
            if (total < 0)total = 0; // Joe Bergin 11/00

            //  int n = _in.read(_ioBuf,0,contentLength);
            //System.out.println(new String(_ioBuf);
           // The original code was too optimistic about grabbing a large block of code.
           // If contentLength is large (large file) you may only grab part of it.
           // Joe Bergin 11/00

            if(total != contentLength)
            {
                doDebug(1, "Amount of content read did not match content length, (" + (total) + ")");
            }

            doDebug(4, "Read content from browser of: " + new String(_ioBuf, 0, total));
            contentLength = total;
        }

        if(_in.available()>0)_in.skip(_in.available()); // flush any remaining stuff

        //Handle the request, any content is sitting in _ioBuf
        if(type.equalsIgnoreCase("GET")) // (Ignores any content in the request.)
        {
            showPage(urls,  null);
        }
        else if(type.equalsIgnoreCase("POST"))
        {
            doPOST(urls, _ioBuf, contentLength);
        }
        else
        {
            doDebug(1, "Unknown HTTP request type: " + type);
            return;
        }

        flush();

    }

    public void showPage(String urls, String password) throws Exception
    {
           File f;
        	if(urls.length() == 0)urls = "WikiIntro";
            f = locateFile(urls);
			String createURL = urls;
			boolean showCreatePage = false;
			if (f==null) // Joe Bergin
			{
				f = new File(_parent.getRoot(), "docs" + File.separator + "noPage");
				showCreatePage = true;
			}
			//else
		//	{	createURL = "";// Don't fill in the create form for an existing page.
		//	}
            FileInputStream fin = new FileInputStream(f);
        	Date lastModified = new Date(f.lastModified());
            Hashtable hdr = readHeader(fin, new Hashtable());

			// Will be none, all, or passwd
			// If no header then the file defaults to no permission.
			// This hides helper files like preamble from being viewed
			// or edited directly.
			String readPerm = (String)hdr.get("read");

			if(readPerm == null || readPerm.equals("none"))
			{
				transmitString("Permission Error","No read permission for that file.");
				try { fin.close(); } catch (Exception nr) { }
				return;
			}

			if(password == null && readPerm.equals("passwd") && !userLoggedIn())
			{
                transmitString("Password Protected", "You must give a password to see that page."+
					"<BR><form method=\"post\" action=\"password\">\n" +
                    "Page:<INPUT TYPE=text name=page VALUE=" + urls +" cols=20><BR>\n" +
                    "<input type=\"Submit\" name=Submit value=\"Password\">\n"  +
					"<input type=\"Hidden\" name=kind value=\"Read\">\n" +
                    "</form>" +
                    "\n&nbsp;The next page will accept your password."

                );

/*				transmitString("Login Error","You must login to see that file." +
					"<BR><form method=\"post\" action=\"login\">\n" +
                    "<input type=\"Submit\" name=Submit value=\"Login\">\n"  +
                    "</form>" +
                    "\n&nbsp;Login here then Back to this page and press Reload");

*/
				try { fin.close(); } catch (Exception nr2) { }
				return;
			}

            String pagePassword = (String)hdr.get("rpass");

            if(pagePassword == null || pagePassword.equals("") ||userLoggedIn() || (password != null && password.equals(pagePassword) ))
            {
                // Simply get the requested page
                transmitHtmlPage(hdr, urls, showCreatePage, fin, lastModified, f.length());
            }
            else
            { // wrong password && not logged in.
                transmitString("Incorrect Password", "The entered password was not correct for "+ urls );
            }


			try { fin.close(); } catch (Exception nr2) { }
    }

	/** Returns false on not authorized.
	 ** Login is detected by the cookie WikiLogin being set to the user name.
	 **/
	public boolean userLoggedIn()
	{
		return _userLogin;
	}

	/** Tries to get a lock on a file.  Returns true if you got it, false
	 ** if already locked.
	 **/
	public boolean getEditLock(String fileName)
	{
		return _parent.getEditLock(fileName);
	}

	/** Releases the lock to edit a file. Is not an error if you don't have
	 ** a lock.
	 **/
	public void releaseEditLock(String fileName)
	{
		_parent.releaseEditLock(fileName);
		return;
	}


    /** A POST is handled by assuming the urls is the class name that handles the
     ** POST. For example, if urls is "edit" then jwiki.edit is the class name.
     ** A class to do this must implement the WikiPostable interface.
     **/
    public void doPOST(String cname, byte buf[], int bufSize)
    {
        Class c;
        WikiPostable p;

        try
        {
            c = Class.forName("jwiki." + cname);
        }
        catch (Exception ec)
        {
            doDebug(1, "Class not found for POST failed, jwiki." + cname);
            return;
        }

        try
        {
            p = (WikiPostable)c.newInstance();
        }
        catch (Exception ec1)
        {
            doDebug(1, "Class new instance for POST failed, jwiki." + cname);
            return;
        }

        try
        {
            p.handlePOST(this, buf, bufSize);
        }
        catch (Exception ec2)
        {
            doDebug(1, "jwiki." + cname + " Failed, " + ec2.getMessage());
            return;
        }

    }

    /** Parses a buffer of POST data into entries in a Hashtable.
     ** @param buf The buffer of data.
     ** @param n How much data is in buf.
     ** @returns The new table, or null on error.
     **/
    public Hashtable parsePOST(byte buf[], int n)
    {
        Hashtable h = new Hashtable();

        String key;
        String val;
        StringBuffer tmp = new StringBuffer(_maxBuf);

        int i = 0;

        for(i=0; i<n; i++)
        {
            if(buf[i] == '%')
            {
                i++;
                byte b;
                if(buf[i] >= '0' && buf[i] <= '9')
                {
                    b = (byte)(buf[i] - '0');
                }
                else if(buf[i] >= 'a' && buf[i] <= 'f')
                {
                    b = (byte)(10 + buf[i] - 'a');
                }
                else
                {
                    b = (byte)(10 + buf[i] - 'A');
                }
                b = (byte)(b * 16);

                i++;
                if(buf[i] >= '0' && buf[i] <= '9')
                {
                    b += (byte)(buf[i] - '0');
                }
                else if(buf[i] >= 'a' && buf[i] <= 'f')
                {
                    b += (byte)(10 + buf[i] - 'a');
                }
                else
                {
                    b += (byte)(10 + buf[i] - 'A');
                }
                tmp.append((char)b);
            }
            else if(buf[i] == '&')
            {
                String ln = tmp.toString();
                doDebug(2, "parsePOST: Got " + ln);
                int idx = ln.indexOf("=");
                if(idx < 0)
                {
                    return null;
                }
                key = ln.substring(0, idx);
                val = ln.substring(idx + 1);
                doDebug(2, "parsePOST: Key:" + key + "  value:" + val);

                h.put(key, val);
                tmp.setLength(0);
            }
            else if(buf[i] == '+')
            {
                tmp.append(' ');
            }
            else
            {
                tmp.append((char)(buf[i]));
            }
        }

        String lnl = tmp.toString();
        doDebug(2, "parsePOST: Got " + lnl);
        int idxl = lnl.indexOf("=");
        if(idxl < 0)
        {
            return null;
        }
        key = lnl.substring(0, idxl);
        val = lnl.substring(idxl + 1);
        doDebug(2, "parsePOST: Key:" + key + "  value:" + val);
        h.put(key, val);

        return h;
    }

    /** Transmits a message or arbitrary String as a page instead of
     ** sending a file .  Sticks a &lt;PRE&gt; block around the string.
     **/
    public void transmitString(String title, String contents)
        throws IOException
    {
        sendString("HTTP/1.0 200 Ok");
        crlf();
        // Headers
        // This keeps the browser from expecting a persistant connection.
        sendString("Connection: close");
        crlf();

        crlf();
        flush();

        startPage(title, false);

        contents = " \n" + contents + "\n</PRE>";
        doDebug(1, "Writing contents:" + contents);
        sendString(contents);

        endPage();

        flush();
    }

    /** Transmits the JWiki page called pageName  Puts the preamble and postamble
     ** and such in.
     **/
    public void transmitHtmlPage(Hashtable hdr, String pageName, boolean showCreatePage, FileInputStream fin, Date lastModified, long size)
        throws IOException
    {
        sendString("HTTP/1.0 200 Ok");
        crlf();
        // Headers
        sendString("Connection: close");
        crlf();

        crlf();
        flush();

        // Do the preamble if needed.
		boolean needHeader = hdr.get("preamble") != null && ((String)hdr.get("preamble")).equalsIgnoreCase("true");
		String hasEditButton = (String)hdr.get("editbutton");
        if(needHeader)
        {	startPage(pageName, false);
        }
		else
		{	writeHeadline(pageName, false);
		}
		if(hasEditButton != null && hasEditButton.equalsIgnoreCase("true"))
		{	sendBlindEditForm(pageName," Last edited on: " + lastModified);
		}
        String contents = getFile(fin);
        if(contents == null)
        {
            doDebug(1, "File transfer failed (1)");
        }

        if(hdr.get("html") == null && ((String)hdr.get("html")).equalsIgnoreCase("false"))
        {
            contents = safeHTML(contents);
        }

        if(hdr.get("wiki") != null && ((String)hdr.get("wiki")).equalsIgnoreCase("true"))
        {
            contents = wikiToHTML(contents);
        }
        if(!JWiki.mayIndex())sendString("<META NAME=\"ROBOTS\" CONTENT=\"NOINDEX, NOFOLLOW, NOARCHIVE\"> <BR>");
        sendString(contents);
        sendString("<HR>" );

        if(hasEditButton != null && hasEditButton.equalsIgnoreCase("true"))
        {
            sendEditForm(pageName, showCreatePage);
        }
		if(userLoggedIn() || allMayDeleteRename())
		{   sendDeletePageForm(pageName);
			sendRenamePageForm(pageName);
		}
        sendSearchForm(pageName);
        sendLoginForm();
        sendString("Page last updated on: " + lastModified); // Joe Bergin
		sendString("<br>Page size approximately: " + size + " bytes.");


        if(hdr.get("postamble") != null && ((String)hdr.get("postamble")).equalsIgnoreCase("true"))
        {	endPage();
        }
        else if(needHeader)
        {	minimumEndPage();
        }

        flush();
    }

    public void sendLoginForm() throws IOException
    {
   		 sendString(
                "<form method=\"post\" action=\"login\">\n" +
                "<input type=\"Submit\" name=Submit value=\"Login\">\n"  +
                "</form>\n");
    }

	public void sendCreateForm(String fileName) throws IOException
	{	sendString("\n" +
            "<form method=\"post\" action=\"edit\">\n" +
        	"<input type=\"Submit\" name=Submit value=\"Create New Page\">\n" +
       		"<INPUT TYPE=text NAME=\"file\" VALUE=\"" + fileName + "\"> " +
			"A new page name should be a <a href=http:./BumpyWord> BumpyWord . </a>\n" +
        	"</form>  \n");
	}

    public void sendEditForm(String fileName, boolean createPage) throws IOException
    {
		if (createPage)
		{	sendCreateForm(fileName);
			return;
		}
    	sendString("\n" +
            "<form method=\"post\" action=\"edit\">\n" +
            "<input type=\"Submit\" name=Submit value=\"Edit\">\n" );
       // if(createPage)
		//	sendString("<input type=\"Submit\" name=Submit value=\"Create New Page\">\n" );
        sendString(	"<INPUT TYPE=text NAME=\"file\" VALUE=\"" + fileName + "\"> ");
		//if(createPage)
		//	sendString("A new page name should be a <a href=http:./BumpyWord> BumpyWord . </a>\n" );
        sendString("</form>  \n");    // Joe Bergin
		//if(!createPage)sendCreateForm();
    }

	public void sendBlindEditForm(String fileName,String message) throws IOException
	{
		 sendString("\n" +
			 "<form method=\"post\" action=\"edit\">\n" +
			 "<input type=\"Submit\" name=Submit value=\"Edit\">" + message +"\n" );
		 sendString(	"<INPUT TYPE=\"Hidden\" NAME=\"file\" VALUE=\"" + fileName + "\"> ");
		 sendString("</form>  \n");    // Joe Bergin
	}

    public void sendSearchForm(String pageName) throws IOException
    {
    	sendString(
    		"<FORM METHOD=\"post\" ACTION=\"search\">\n" +
    		"<INPUT TYPE=\"Submit\" VALUE=\"Search\">\n" +
    		"<INPUT TYPE=text Name=\"whatfor\" Value=\""+ pageName + "\">\n" +
    		"</FORM>\n");
    }

	public void sendRenamePageForm(String pageName)throws IOException
	{   sendString(
		"<FORM METHOD =\"post\" ACTION=\"renamedone\"> \n" +
		"<INPUT TYPE=\"Submit\" VALUE=\"Rename Page\">  " +
		"<INPUT TYPE=\"Hidden\" NAME=\"oldname\" VALUE=\"" + pageName + "\">\n" +
		"<INPUT TYPE=text Name=\"newname\" Value=\""+ pageName + "\">\n" +
		"</FORM>\n"
		);
	}

	public void sendDeletePageForm(String pageName) throws IOException
	{   sendString(
		"<FORM METHOD =\"post\" ACTION=\"deletedone\"> \n" +
		"<INPUT TYPE=\"Submit\" VALUE=\"Delete Page\">\n" +
		"<INPUT TYPE=\"Hidden\" NAME=\"name\" VALUE=\"" + pageName + "\">\n" +
		"</FORM>"
		);
	}

    /** Writes the string out.  This is encapsulated in case we sant
     ** to change how writes are done later.
     **/
    public void sendString(String contents)
        throws IOException
    {
        _out.write(contents.getBytes());
    }


    /** Reads header till double crlf.
	 ** Can be used to read a HTTP header or a file header in the docs directory.<BR>
	 **
     ** The resulting Hashtable is where key is a String of the header name,
	 ** and value is a String of the header field value.  Does not concatenate
	 ** multiple header instances. (yet).
	 ** cookies is a Hashtable where the values of any found cookies are returned.
	 ** The key is a String with the cookie name, and the value is a String
	 ** with the rest of the cookie info. We fake the parsing.  This is not
	 ** full featured and robust cookie pareing.  It only looks for the word
	 ** WikiLogin.
     **/
    public Hashtable readHeader(InputStream in, Hashtable cookies)
    {
        Hashtable h = new Hashtable();
		cookies.clear();

        String s = readLine(in);
        while(s.length() != 0)
        {
            int idx = s.indexOf(":");
            if(idx == -1)
                return h;
            String key = s.substring(0, idx );
            String val = s.substring(idx + 1);
			doDebug(2, "readHeader: " + s);
			if(key.equalsIgnoreCase("cookie"))
			{
				doDebug(2, "Got cookie line: " + val);
				int cidx = val.indexOf("WikiLogin=true");
				if(cidx > 0)
				{
				   cookies.put("wikilogin", "true");
				}
			}
            doDebug(3, "Putting file header   key:" + key + "  value:" + val);
            h.put(key, val);
            s = readLine(in);
        }

        return h;
    }

    /** Finds the given file by name.  Searches in the docs area.
     ** @param pageName The name of the page.  Pages usualy do not have dot endings.
     ** @throws Exception Can throw if no file.
     **/
    public File locateFile(String pageName)
        throws Exception
    {
		File f;

     /*   if (pageName.length()==0)
        {	f = new File(_parent.getRoot(), "docs" + File.separator + "WikiIntro");
    		return f;
        }
*/
		// Don't allow hackers to access the file system out of the docs directory.
		if(pageName.length() == 0 || // Must not be 0 length
		   pageName.indexOf("..") != -1 ||  // Don't do ../  or ..\
		   pageName.indexOf("/") != -1 ||
		   pageName.indexOf("\\") != -1 ||
		   pageName.charAt(0) == '.') // Avoid hidden unix files and bogus names.
		{
            f = new File(_parent.getRoot(), "docs" + File.separator + "noPage");
			return f;
		}


        f = new File(_parent.getRoot(), "docs" + File.separator + pageName);
        if(! f.exists())
        {	f = null; // Joe Bergin Want to avoid filling in create page form for existing page.
            //f = new File(_parent.getRoot(), "docs" + File.separator + "noPage");//JB 9/2001
        }
        doDebug(3, "locatePage returning " + f);
        return f;
    }


    /** Kills the thread of this JHandler. Causes the JHandler's thread to terminate soon.
     **/
    public void kill()
    {
        /** This should wake up any pending IO on the socket. **/
        try { _sock.close(); } catch (Exception e) { }
    }

    /** Emits the standard crlf sequence for HTTP regardless of the platform
     ** we are on.
     **/
    void crlf()
        throws IOException
    {
        _out.write(_crlf, 0, 2);
    }

    void crlf(OutputStream f)
        throws IOException
    {
        f.write(_crlf, 0, 2);
    }


    /** Simply _out.flush(); **/
    void flush()
        throws IOException
    {
        _out.flush();
    }

    /** Copy data from fin to _out till fin hits end of file.
     ** @returns true on success.
     **/
    boolean transferFile(FileInputStream fin)
    {
        doDebug(2, "Transferring file (given FileInputStream)");
        try
        {

            int n = fin.read(_ioBuf);
            while(n > 0)
            {
                _out.write(_ioBuf, 0, n);
                flush();
                n = fin.read(_ioBuf);
            }

            fin.close();
         }
         catch (Exception e)
         {
            return false;
         }

        return true;
    }

    String getFile(FileInputStream fin)
    {
        doDebug(2, "Getting file (given FileInputStream)");
        StringBuffer b = new StringBuffer();
        try
        {
            int n = fin.read(_ioBuf);
            while(n > 0)
            {
                b.append(new String(_ioBuf, 0, n));
                n = fin.read(_ioBuf);
            }
                     }
         catch (Exception e)
         {
            return null;
         }

        return b.toString();
    }


    /** Converts s such that any HTML special characters are encoded as
     ** ampersand-whatever-semicolon
     **/
    String safeHTML(String s)
    {
        StringBuffer b = new StringBuffer();
        int i;

        doDebug(2, "safeHTML(String)");
        for(i=0; i<s.length(); i++)
        {
            int c = s.charAt(i);
            switch(c)
            {
                case '"':
                    b.append("&quot;");
                    break;
                case '&':
                    b.append("&amp;");
                    break;
                case '<':
                    b.append("&lt;");
                    break;
                case '>':
                    b.append("&gt;");
                    break;
                default:
                    b.append((char)c);
                    break;
            }
        }
        return b.toString();
    }

    /** Converts string to HTML assuming the contents are Wiki style text.
     **/
    String wikiToHTML(String s)
		throws IOException
    {
    	// This is all in the JWikiCvt class since it's rather large.
        return JWikiCvt.wikiToHTML(s);
    }

    boolean transferFile(File f)
    {
        FileInputStream fin = null;
        boolean ret;

        doDebug(2, "Transferring file " + f);
        try
        {
            fin = new FileInputStream(f);
            ret = transferFile(fin);
        }
        catch (Exception e)
        {
            try { fin.close(); } catch(Exception ec) { }
            return false;
        }


        // Bullet proof close
        try { fin.close(); } catch(Exception ec2) { }

        return true;
    }

       /** Opens and reads a file into a String and returns it.
        ** Retirns null on error.
        **/
    public String getFile(File f)
    {
        FileInputStream fin = null;
        String ret;

        try
        {
            fin = new FileInputStream(f);
            ret = getFile(fin);
        }
        catch (Exception e)
        {
            try { fin.close(); } catch(Exception ec) { }
            return null;
        }


        // Bullet proof close
        try { fin.close(); } catch(Exception ec2) { }

        return ret;
    }

    /** Reads a line up to the terminating crlf and puts it in the _ioBuf
     ** If _ioBuf[0] is '\0' then end of header lines.
     ** Returns a String or "" if no more data.
     **/
    String readLine(InputStream in)
    {
        int i = 0;
        int n;

        try
        {
            n = in.read(_ioBuf, i, 1);
            //doDebug(4, "readLine got " + n + " bytes");

            while(n > 0 && _ioBuf[i] != LF && i < (_ioBuf.length - 1))
            {
                if(_ioBuf[i] != CR)
                {
                    i++;
                }
               n =  in.read(_ioBuf, i, 1);
               //doDebug(4, "readLine got " + n + " bytes");
            }
        }
        catch(Exception e)
        {
            System.out.println("exception in JHandler:readLine: " + e.getMessage());
            e.printStackTrace();
        }

        String s = new String(_ioBuf, 0, i);
        doDebug(4, "readLine [" + s + "]");
        return s;
    }

	public void transmitEditPage(String fileName, String password) throws IOException
	{
		Hashtable cookies = new Hashtable();
		JHandler h = this;
		FileInputStream f = null;
		if(!h.getEditLock(fileName))
		{
			h.transmitString("Conflict Error","Someone is editing the page " + fileName + ". (Or the file does not exist.)");
			return;
		}

		try
		{
			f = new FileInputStream(
		    	h.getParent().getRoot() + File.separator +
		    	"docs" + File.separator + fileName);
		}
		catch(Exception oe)
		{
			h.transmitString("Page Not Found Error", "Can not open page.");
		    h.releaseEditLock(fileName);
		 	return;
		}

		Hashtable fhdr = h.readHeader(f, cookies);


        //OutputStream out = h.getOut();
        h.sendString("HTTP/1.0 200 Ok");
        h.crlf();
        // Headers
        // This keeps the browser from expecting a persistant connection.
        h.sendString("Connection: close");
        h.crlf();
        h.crlf();

        h.startPage("Edit of Page: "+ fileName,false);

        h.sendString("Make sure you exit this edit page with the <B>Save</B>\n");
        h.sendString("or <B>Cancel</B> button. \n Otherwise the page lock won't be released.");
        h.sendString("Wait for the entire page to load.");

		String readPerm = (String)fhdr.get("read");
		String writePerm = (String)fhdr.get("write");
		String rPassword = (String)fhdr.get("rpass");
		String wPassword = (String)fhdr.get("wpass");

		if(writePerm == null || writePerm.equals("none"))
		{
			try { f.close(); } catch (Exception fe) { }
            h.sendString("\n<h3>You do not have permission to edit that page.</h3>");
            h.endPage();
            h.releaseEditLock(fileName);
            return;
		}

		if(writePerm.equals("passwd") && !h.userLoggedIn() && !password.equals(wPassword))
		{
			try { f.close(); } catch (Exception fe) { }
            h.sendString("\n<h3>Password protected: You must supply a password to edit that page.</h3>\n");
 //           h.sendLoginForm();
 			h.sendString("<BR><form method=\"post\" action=\"password\">\n" +
 			"Page:<INPUT TYPE=text name=page VALUE=" + fileName +" cols=20><BR>\n" +
 			"<input type=\"Submit\" name=Submit value=\"Password\">\n"  +
 			"<input type=\"Hidden\" name=kind value=\"Write\">\n" +
 			"</form>" +
 			"\n&nbsp;The next page will accept your password."
 			);
            h.endPage();
            h.releaseEditLock(fileName);
            return;
		}

        // Now the body of the form.
        /* IMPORTANT NOTE: The save button was removed from the top of a page. The
        	remaining button is at the bottom. With a button at the top it was possible to
            save the page before the edit form completely loaded. The POST would then be
            incomplete. This resulted in truncated files being saved which meant that
            either you lost content or (worse) you lost the page characteristics. The file
            would then become corupted and the page became unusable.
        */
        h.sendString("<FORM METHOD=POST ACTION=editsave>\n");
 //       h.sendString("<INPUT TYPE=\"Submit\" NAME=Submit VALUE=\"Save\">\n");
        h.sendString("<INPUT TYPE=\"Submit\" NAME=Submit VALUE=\"Cancel\"><P>\n");
        h.sendString("<TEXTAREA name=txt cols=80 rows=15>\n");

        String contents = h.getFile(f);

		try { f.close(); } catch (Exception fe) { }

        contents = h.safeHTML(contents);
        h.sendString(contents);

        h.sendString("</TEXTAREA><BR>\n");
        h.sendString("<INPUT TYPE=HIDDEN NAME=\"file\" VALUE=\"" + fileName + "\">\n");

        String temp = (String)fhdr.get("html");
        if(temp != null && temp.equalsIgnoreCase("TRUE"))
            h.sendString("Allow HTML:<INPUT TYPE=checkbox CHECKED NAME=html VALUE=\"html\">\n");
        else
            h.sendString("Allow HTML:<INPUT TYPE=checkbox NAME=html VALUE=\"html\">\n");

        h.sendString("<BR>");
        temp = (String)fhdr.get("wiki");
        if(temp != null && temp.equalsIgnoreCase("TRUE"))
            h.sendString("Wiki:<INPUT TYPE=checkbox CHECKED NAME=wiki VALUE=\"wiki\">\n");
        else
            h.sendString("Wiki:<INPUT TYPE=checkbox NAME=wiki VALUE=\"wiki\">\n");

        h.sendString("<BR>");
        temp = (String)fhdr.get("preamble");
        if(temp != null && temp.equalsIgnoreCase("TRUE"))
            h.sendString("Preamble:<INPUT TYPE=checkbox CHECKED NAME=preamble VALUE=preamble>\n");
        else
            h.sendString("Preamble:<INPUT TYPE=checkbox NAME=preamble VALUE=preamble>\n");

        h.sendString("<BR>");
        temp = (String)fhdr.get("postamble");
        if(temp != null && temp.equalsIgnoreCase("TRUE"))
            h.sendString("Postamble:<INPUT TYPE=checkbox CHECKED NAME=postamble VALUE=postamble>\n");
        else
            h.sendString("Postamble:<INPUT TYPE=checkbox NAME=postamble VALUE=postamble>\n");

        h.sendString("<BR>");
        temp = (String)fhdr.get("editbutton");
        if (temp != null && temp.equalsIgnoreCase("TRUE"))
        {   h.sendString("Show edit button:<INPUT TYPE=checkbox CHECKED NAME=editbutton VALUE=editbutton>\n");
       	}
        else
             h.sendString("Show edit button:<INPUT TYPE=checkbox NAME=editbutton VALUE=editbutton>\n");
        h.sendString("<B>Warning:</B> Once you turn this off you can only turn it on by editing the file directly on the server machine.<BR>\n");

		if(writePerm == null || writePerm.equals("all"))
            h.sendString("Write Permission:<INPUT TYPE=radio CHECKED NAME=writePerm VALUE=all>all&nbsp;\n");
		else
            h.sendString("Write Permission:<INPUT TYPE=radio NAME=writePerm VALUE=all>all&nbsp;\n");

		if(writePerm.equals("passwd"))
            h.sendString("<INPUT TYPE=radio CHECKED NAME=writePerm VALUE=passwd>password required&nbsp;\n");
		else
            h.sendString("<INPUT TYPE=radio NAME=writePerm VALUE=passwd>password required&nbsp;\n");
        h.sendString("|| WritePassword ");
		if(wPassword == null || wPassword.equals(""))
			h.sendString("<INPUT TYPE=text name=wpass cols=20> \n");
		else
			h.sendString("<INPUT TYPE=text name=rpass value="+wPassword+" cols=20> \n");

		if(writePerm.equals("none"))
            h.sendString("<INPUT TYPE=radio CHECKED NAME=writePerm VALUE=none>not writeable <BR>");
		else
            h.sendString("<INPUT TYPE=radio NAME=writePerm VALUE=none>not writeable <BR>");
		if(readPerm == null || readPerm.equals("all"))
            h.sendString("Read Permission:<INPUT TYPE=radio CHECKED NAME=readPerm VALUE=all>all&nbsp;\n");
		else
            h.sendString("Read Permission:<INPUT TYPE=radio NAME=readPerm VALUE=all>all&nbsp;\n");

		if(readPerm.equals("passwd"))
            h.sendString("<INPUT TYPE=radio CHECKED NAME=readPerm VALUE=passwd>password required ");
		else
            h.sendString("<INPUT TYPE=radio NAME=readPerm VALUE=passwd>password required ");

        h.sendString("|| Read Password ");
        if(rPassword==null || rPassword.equals(""))
            h.sendString("<INPUT TYPE=text name=rpass cols=20><BR>\n");
        else
            h.sendString("<INPUT TYPE=text name=rpass value="+rPassword+" cols=20><BR>\n");
        h.sendString("&nbsp;&nbsp;&nbsp;<b>Warning:</b> Page passwords, if required, must be non empty.<BR>\n");

        h.sendString("<P>");

        h.sendString("<INPUT TYPE=\"Submit\" NAME=Submit VALUE=\"Save\">\n");
        h.sendString("<INPUT TYPE=\"Submit\" NAME=Submit VALUE=\"Cancel\">\n");

        h.sendString("</FORM>\n");

        h.endPage();
        h.flush();

	}

    /** Calls the _parent doDebug **/
    public void doDebug(int lev, String mes)
    {
        _parent.doDebug(lev, mes);
    }

}
