package jwiki;

import java.io.*;
import java.util.*;
import java.net.*;

/** The JWiki server.<BR>
 ** This server runs with no awt window.  It listens on port 8088 (default)
 ** The command line argument is the root of the JWiki file system.<BR>
 ** In the root are the following files:<BR>
 ** <UL>
 ** <LI> log.txt  The log file
 ** <LI> docs The directory of doccuments.  Each doccument is a file.
 ** <LI> config.ini The configuration INI file.  This file has various tags
 ** that configure JWiki.
 ** <LI> passwords The password file.  Simply is name password pairs sparated by a :
 ** <LI> password.html The body of the response to edit your password.
 ** </UL> <BR>
 ** There may be other files as I think of them.<P>
 ** The config.ini file is generaly tag:value entries.  You may have comments
 ** that start with # if they are aline on the line.  You may have blank lines.
 ** Tags are:
 ** <UL>
 ** <LI> port:portNumber
 ** <LI> secure:true or false
 ** <LI> root:root of the file system. Contains directory docs.
 ** <LI> recent: number of days worth of recent changes to list
 ** <LI> logfile:file specification of log file
 ** <LI> passwordfile:file spec of password file
 ** <LI> spelling: file that contains the spell dictionary HashSet
 ** <LI> localspell: file that contains the local extensions (text file)
 ** <LI> debug:debug level 0 = off.
 ** <LI> edittimeout:minutes until a lock expires
 ** <LI> webgif:url of a graphic to display on each page
 ** <LI> index: true or false. Permit web spiders to index the pages
 ** <LI> prefix: String used to prefix page name in page titles
 ** <LI> html:true or false
 ** <LI> wiki:true or false
 ** <LI> preamble:true or false
 ** <LI> postamble: true or false
 ** <LI> editbutton:true or false
 ** <LI> read:all, passwd, or none
 ** <LI> write:all, passwd, or none
 ** <LI> adminpasswd: password used (with root username) to kill system
 ** <LI> allmaydelete:true or false, anyone may delete/rename pages if true 
 *			otherwise requires login.
 ** </UL>
 **	<br>
 **  This was written with Java 1.1.7 to permit maximum portability. It is suggested
 ** that others extending this keep the same convention.
 **
 ** Note that some of the classes purposly have names beginning with lower case letters
 ** contrary to the usual Java coding conventions. These classes are the keys used in POST
 ** messages and are used with the Java reflection API to handle POST requests.
 **/

public class JWiki implements Runnable {

	private static int _port;
	static JWiki myself = null;
	private File _root;
	private PrintWriter _logFile;
	private int _debugLevel;
	private boolean _secure = true;
	private boolean _keepRunning;

	/** The list of all spawned handlers. **/
	private Vector _handlers;
	private ServerSocket _sock;

	/** A Hashtable of files that are locked.  The key is the file name, the
	 ** value is the Date it was locked. **/
	private Hashtable _editLocks;
	private static long _editTimeout;
	//The passwords table.  key is the user name as a String, value is the
	// plaintext password as a String.
	private Hashtable _passwords;
    private static String _passwordFileName = "passwords";
    private static String _spellDictionaryName ="words.hashset";
    private static String _localDictionaryName ="localwords.txt";
    private static boolean _mayIndex = false;
	private static boolean _allMayDeleteRename = false;
	private static String titlePrefix = "";

    // The url of a graphic to display on each page. From config.ini
    private static String webGifURL = null;

    // cached preamble and postamble
    private static String preambleFirst = null; // JoeBergin 11/00
    private static String preambleRest = null;
    private static String postambleAll = null;
    private static String postambleMinimum = "</BODY> </HTML>";

    // default permissions for new pages. Reset from config file.
    private static String allowHtml = "true";
    private static String allowWiki = "true";
    private static String usePreamble = "true";
    private static String usePostamble = "true";
    private static String showEditButton = "true";
    private static String readPermission = "all";
    private static String writePermission = "all";
	private static String adminpw = null;
    private static int recent = 1;

	private static Thread driverThread;

    static String preambleFirst()
    {	return preambleFirst;
    }

    static int recent()
    {	return recent;
    }

    static boolean mayIndex()
    {	return _mayIndex;
    }

	static boolean allMayDeleteRename()
	{   return _allMayDeleteRename;
	}

    static int port()
    {	return _port;
    }

    static String preambleRest()
    {	return preambleRest;
    }

    static String postambleAll()
    {	return postambleAll;
    }

    static String postambleMinimum()
    {	return postambleMinimum;
    }

    static String allowHtml()
    {	return allowHtml;
    }

    static String allowWiki()
    {	return allowWiki;
    }

    static String usePreamble()
    {	return usePreamble;
    }

    static String usePostamble()
    {	return usePostamble;
    }

    static String showEditButton()
    {	return showEditButton;
    }
    
    static String titlePrefix()
    {	return titlePrefix;
    }
    
    static String spellDictionary()
    {	return _spellDictionaryName;
    }
    
    static String localDictionary()
    {	return _localDictionaryName;
    }

    static String readPermission()
    {	return readPermission;
    }

    static String writePermission()
    {	return writePermission;
    }

	static String adminpw()
	{	return adminpw;
	}

	public File getRoot() {
		return _root;
	}

	/** @param root The root of the JWiki file system.
    **  @param logfile Unused in this version. Will permit setting logfile from command
    ** line instead of config file (with other changes).
	**/
	public JWiki(String root, PrintWriter logfile) {  // Joe Bergin 11/00
		_debugLevel = 0;
		_editTimeout = 5; // Five minutes default.
		_secure = false;
		_logFile = logfile;
		_handlers = new Vector(10, 10);
		_sock = null;
		_keepRunning = true;

        File configRoot = new File(root);
        if(configRoot == null) configRoot = new File(".");
        File configFile = new File(configRoot, "config.ini");

		processIniFile(configFile);
        try
        {	getPrePostamble();// Joe Bergin 11/00

        }catch(Exception fex)
        {	doDebug(0, "Failed getting Pre and Postamble files");
        }

		_editLocks = new Hashtable();

		_passwords = new Hashtable();
		readPasswords();
		myself = this;
	}

    public void getPrePostamble() throws FileNotFoundException, IOException  	   // Joe Bergin 11/00
    {	File preambleFile = new File(getRoot(), "docs" + File.separator + "preamble1");
    	File preambleCompletion = new File(getRoot(),"docs" + File.separator + "preamble2");
    	File postambleFile = new File(getRoot(), "docs" + File.separator + "postamble");
    	BufferedInputStream fin = new BufferedInputStream(new FileInputStream(preambleFile));
        int length = fin.available();
        byte [] buff = new byte[length];
        int n = fin.read(buff);
        preambleFirst = new String(buff);

        fin = new BufferedInputStream(new FileInputStream(preambleCompletion));
        length = fin.available();
        buff = new byte[length];
        n = fin.read(buff);
        preambleRest = new String(buff);

        fin = new BufferedInputStream(new FileInputStream(postambleFile));
        length = fin.available();
        buff = new byte[length];
        n = fin.read(buff);
        postambleAll = new String(buff);

    }

	/** Reads in the config.ini file setting paramaters.
	**/
	public void processIniFile(File f) {
		BufferedReader in;

	//	File f = new File(_root, "config.ini");
		if(f.exists() == false) {
			doDebug(0, "No config.ini file in " + _root);
			return;
		}
		try {
			in = new BufferedReader(new FileReader(f));

			String s = in.readLine();
			while(s != null) {
				s = s.trim();

				doDebug(2, "config.ini: " + s);

				// Not a comment line and not a blank line.
				if( s.length() != 0 && s.charAt(0) != '#' ) {
					String tag;
					String val;

					int i = s.indexOf(":");
					if(i == -1) {
						doDebug(0, "Bad line in config.ini: " + s);
						return;
					}
					tag = s.substring(0, i).toUpperCase();
					val = s.substring(i + 1).trim();
					if (tag.equals("ROOT"))
                    {	_root = new File(val);
                    	if(_root.isDirectory() == false) {
                    		doDebug(0, "The root directory does not exist. [" + val + "]");
                    		System.exit(1);
                    	}

                    }else
					if(tag.equals("ALLMAYDELETE"))
					{   if(val.equalsIgnoreCase("true"))
						{   _allMayDeleteRename = true;
						}
					}
					else
                    if (tag.equals("LOGFILE"))
                    {	_logFile = new PrintWriter(new FileWriter(val));// Joe Bergin 11/00
                    }
                    else if(tag.equals("PORT")) {
						_port = Integer.parseInt(val);
					} 
					else if (tag.equals("RECENT"))
					{	recent = Integer.parseInt(val);
					}
					else if(tag.equals("PREFIX"))
					{	titlePrefix = val!=null?val:"";
					}
					else if(tag.equals("DEBUG")) {
						_debugLevel = Integer.parseInt(val);
					} else if(tag.equals("EDITTIMEOUT")) {
						_editTimeout = (long) Integer.parseInt(val);
					} else if(tag.equals("SECURE")) {
						if(val.equalsIgnoreCase("false"))
							_secure = false;
						else
							_secure = true;
                    } else if(tag.equals("INDEX")){
                            if(val.equalsIgnoreCase("true"))
                                   _mayIndex = true;
                            else
                                   _mayIndex = false;
					} else if(tag.equals("WEBGIF"))
					{	webGifURL = val;
					} else if(tag.equals("ADMINPASSWD"))
					{	adminpw = val;
					} 
					else if(tag.equals("SPELLING"))
					{	_spellDictionaryName = val;
					}
					else if(tag.equals("LOCALSPELL"))
					{	_localDictionaryName = val;
					}
					else if(tag.equals("HTML")) {
						allowHtml = val;
					} else if(tag.equals("WIKI")) {
						allowWiki = val;
					} else if(tag.equals("PREAMBLE")) {
						usePreamble = val;
					} else if(tag.equals("POSTAMBLE")) {
						usePostamble = val;
					} else if(tag.equals("EDITBUTTON")) {
						showEditButton = val;
                    } else if(tag.equals("READ")) {
                    	readPermission = val;
                    } else if(tag.equals("WRITE")) {
                    	writePermission = val;
                    } else if(tag.equals("PASSWORDFILE"))
                    {	_passwordFileName = val;
                    }
                    else {
						doDebug(0, "Unknown tag " + tag + " in config.ini");
						return;
					}
				}

				s = in.readLine();
			}

		}
		catch (Exception e) {
			System.out.println("exception reading config.ini : " + e.getMessage());
			e.printStackTrace();
			return;
		}

		// Bullet proof close.
		try { in.close();
		} catch (Exception ec) {
		}
	}

	public String webGif()
	{	return webGifURL; // may be null
	}

	public void run() {
		Socket newSock = null;
		JHandler hand;

		doDebug(1, "JWiki thread started.");
		try {
			_sock = new ServerSocket(_port);
		}
		catch (Exception e) {
			System.out.println("exception opening socket : " + e.getMessage());
			e.printStackTrace();
			return;
		}

		while(_keepRunning) {
			try {
				newSock = _sock.accept();

				// This could also be in a seperate thread but that is unnecessary.
				releaseOldEditLocks();

				hand = new JHandler(this, newSock);
				(new Thread(hand)).start();
			}
			catch (Exception ae) {
				System.out.println("exception getting connection : " + ae.getMessage());
				ae.printStackTrace();
			}
		}

		// Bullet proof close
		try { _sock.close();
		} catch(Exception ec) {
		}
	}

	/** Puts mes out to wherever debug goes.  Usualy the log.txt file.
	** @param lev The leve of the message.  Higher levels are more detailed messages.
	** A level of 0 always outputs.
	** @param mes The message to put out.
	**/
	public void doDebug(int lev, String mes) {
		if(_debugLevel < lev)
			return;

		if(_logFile == null) {
			System.out.println(mes);
			return;
		} else // TBD
		{
			_logFile.println(mes);
            _logFile.flush();
			return;
		}
	}

	/** Is true if the given level will cause output. **/
	public boolean willDebug(int lev) {
		return _debugLevel >= lev;
	}

	/** Called by a new handler to put it's self in the _handlers list **/
	public void addHandler(JHandler h) {
		_handlers.addElement(h);
	}

	/** Called by a dead handler to remove it's self from the _handlers list. **/
	public void removeHandler(JHandler h) {
		_handlers.removeElement(h);
	}

	/** Can be called by another thread to terminate this JWiki instance.
	**/
	public void abortWiki() {
		_keepRunning = false;

		// This will wakeup the accept() and make it throw an exception.
		try { _sock.close();
		} catch(Exception e5) {
		}

		for (Enumeration e = _handlers.elements() ; e.hasMoreElements() ;) {
			JHandler h = (JHandler)e.nextElement();
			h.kill();
		}
		// Wait for all to die.
		while(_handlers.size() > 0) {
			try { Thread.sleep(100);
			} catch(Exception ee) {
			}
		}
	}

	/** Tries to get a lock on a file.  Returns true if you got it, false
	 ** if already locked.  Edit locks time out in <tt>_editTimeout</tt>
	 ** minutes.  You can also kill the JWiki server and restart it to
	 ** release all locks.
	 **/
	public synchronized boolean getEditLock(String fileName) {
		for (Enumeration e = _editLocks.keys() ; e.hasMoreElements() ;) {
			String fn = (String)e.nextElement();
			if(fn.equals(fileName)) {
				doDebug(2, "Failed to get edit lock on " + fileName + ". Already locked.");
				return false; // Already locked so return failure.
			}
		}
		_editLocks.put(fileName, new Date());
		doDebug(2, "Got edit lock on " + fileName + ".");
		return true;
	}

	/** Releases the lock to edit a file. **/
	public synchronized void releaseEditLock(String fileName) {
		for (Enumeration e = _editLocks.keys() ; e.hasMoreElements() ;) {
			String fn = (String)e.nextElement();
			if(fn.equals(fileName)) {
				_editLocks.remove(fn);
				doDebug(2, "Removed edit lock on " + fileName + ".");
				return;
			}
		}
		doDebug(2, "Removed edit lock on " + fileName + ". (Was not locked.)");
	}

	/** Releases any old locks on file (if timmed-out). **/
	public synchronized void releaseOldEditLocks() {
		// getTime returns milliseconds from Jan 1 1970
		// The timeout is set in the config.ini file as edittimeout:
		long old = (new Date()).getTime() - 1000 * 60 * _editTimeout;

		for (Enumeration e = _editLocks.keys() ; e.hasMoreElements() ;) {
			String fn = (String)e.nextElement();
			Date dt = (Date)(_editLocks.get(fn));
			if(dt.getTime() <= old) {
				_editLocks.remove(fn);
				doDebug(2, "Removed edit lock on " + fn + ".  Timedout.");
				return;
			}
		}
	}

	/** Returns the edit timeout in minutes. **/
	public int getEditLockTimeout() { return (int)_editTimeout;
	}

	/** Gets the password for the given user.  Returns null on no such user.
	 **/
	public String lookupPassword(String userName) {
		doDebug(2, "lookupPassword(" + userName + ")");
		Object ob = _passwords.get(userName);
		if(ob == null) {
			doDebug(2, "lookupPassword: no such user");
			return null;
		}
		doDebug(2, "lookupPassword: found it.");
		return (String)ob;
	}

	public void readPasswords() {
		_passwords.clear();

		doDebug(2, "readPasswords()");

		// Back door if desired.  Else comment it out.
		//_passwords.put("login", "password");

		// Now read the file called 'passwords'
		readPasswordFile(_passwords);
	}

	private void readPasswordFile(Hashtable passwords) {

		BufferedReader reader = null;
		try {

			reader = new BufferedReader(new FileReader(_passwordFileName));
		} catch (FileNotFoundException e) {
			System.err.println("Password file \"passwords\" not found.  Exiting...");
			System.exit(1);
		}

		StringTokenizer tokenizer;
		String currentLine;

		try {
			while(reader.ready()) {
				currentLine = reader.readLine();
				if (currentLine == null || currentLine.equals(""))
					return;
				tokenizer = new StringTokenizer(currentLine, ":");
				// (pseudo) passwords.put(currentLoginName, currentPassword)
				passwords.put(tokenizer.nextToken(), tokenizer.nextToken());
			}
		} catch (Exception e) {
			System.err.println("Error reading password file.  Exiting...");
			e.printStackTrace();
			System.exit(1);
		}
	}

    public void finalize()
    {_logFile.close();
    }

	/** The only command line arg is not optional and is the root directory
	** of the JWiki file system.
	**/
	public static void main(String[] args) {
                String root;
		if(args.length  < 1) {
			System.out.println("Must have the directory containing config.ini on the command line.");
			//root = ".";
            return;
		}
                else
                    root = args[0];
/*        PrintWriter logfile = null;
        if (args.length > 1)// Joe Bergin 11/00
        {	try
        	{	logfile = new PrintWriter(new FileWriter(args[1]));// Joe Bergin 11/00
        	}
            catch(Exception ex1)
            {
            }
        }
*/		JWiki jWikiServer = new JWiki(root, null);// Joe Bergin 11/00
		driverThread = (new Thread(jWikiServer));
		driverThread.start();
	}

}