package jwiki;

import java.io.*;
import java.util.*;

/** A utility class to convert from Wiki format to HTML.
 ** A --- is converted to the HR tag.  A word with leading
 ** uppercase, then lower case, then uppercase is treated as
 ** a link to the named doccument.
 ** A hyperref like http://www.xmission.com/~rkeene is treated as
 ** a hyper ref.
 ** All other text is unchanged.
 **/
public class JWikiCvt {
    /** Carriage return **/
    private  final static int CR = 0x0D;
    /** Line feed **/ 
    private  final static int LF = 0x0A; 

	// Made from the original input string.
	private StringReader _in;

	private StringBuffer _sBuf;
	
	// Spell check -- Joe Bergin
	
	private static HashSet wordSet = null;
	
	static // Try to initialize the spelling table 
	{	try
		{	ObjectInputStream s = new ObjectInputStream(new FileInputStream(JWiki.spellDictionary()));
			wordSet = (HashSet) s.readObject();
			//System.out.println(""+wordSet.size());
			File docs = new File(JWiki.myself.getRoot() + File.separator + "docs");
			String fnames[] = docs.list();
			for(int i = 0; i < fnames.length; ++i)
			{	wordSet.add(fnames[i]);
			}
		}
		catch(Exception e)
		{ System.out.println("No Spell Checking");
		}
		try
		{	BufferedReader in = new BufferedReader(new FileReader(JWiki.localDictionary()));
			String value = in.readLine();
			while (value != null)
			{	if(!(value.equals("")) && value.charAt(0) != '#')
				{	wordSet.add(value.toUpperCase());
				}
				value = in.readLine();	
			}
		}
		catch(Exception f)
		{	System.out.println("No Local Words");
		}
	}

	private boolean _ungetIsValid;
	private int _ungetChar;

	public JWikiCvt()
	{
		_ungetIsValid = false;
	}
    
    /** Converts string to HTML assuming the contents are Wiki style text.
     **/
    public static String wikiToHTML(String s) throws IOException
    {
		JWikiCvt c = new JWikiCvt();
		return c.wikiToHTMLEntry(s);
    }
	
	// Spell Check 
	
	static void addWord(String word)
	{	if(wordSet != null)	
		{	wordSet.add(word);
		}
	}
    
    static ArrayList spellCheck(String s)throws IOException
    {	if(wordSet == null) return null;
    	JWikiCvt c = new JWikiCvt();  
        return c.spellCheckEntry(s);
    }
	
	private ArrayList spellCheckEntry(String s) throws IOException
	{	ArrayList result = new ArrayList();
		HashSet resultSet = new HashSet();
		
		// Set for one loop when a CR is encountered.
		boolean justCR = true;

		_in = new StringReader(s);

		while(true)
		{
			String frag = readFrag();
			
			if(frag == null)
				break; // Out of while loop.
			if(justCR && frag.indexOf("---") == 0)
			{
			} 
			else if(frag.charAt(0) == CR && justCR)
			{
			}
			else if (frag.charAt(0) == CR || frag.charAt(0) == LF)
			{
			}
			else if( frag.indexOf("http:") == 0 ||
            		 frag.indexOf("shttp:") == 0 ||
					 frag.indexOf("mailto:") == 0 ||
					 frag.indexOf("news:") == 0 ||
					 frag.indexOf("ftp:") == 0 ||
					 frag.indexOf("mailto:") == 0 ||
					 frag.indexOf("gopher:") == 0 ||
					 frag.indexOf("telnet:") == 0)
			{
			}
			else if(isWikiLink(frag))
			{
				if (!wordSet.contains(frag))
				{	resultSet.add(frag);
				}
			}
			
			else if (trimPunctuation(frag).trim().equals(""))
			{
			}
			else
			{	String test = trimPunctuation(frag);	
				if (test.equals(frag) && !wordSet.contains(frag.toUpperCase()))
				{	resultSet.add(frag);
				}
			}

			if(frag.charAt(0) == CR)
			{
				justCR = true;
			}
			else
			{
				justCR = false;
			}
				
		}

		Iterator words = resultSet.iterator();
		while (words.hasNext())
		{	result.add(words.next());
		}
		Collections.sort(result);
	
		return result;
	}

	String trimPunctuation(String s)
	{	String result= "";
		for(int i = 0; i < s.length(); ++i)
		{	if(Character.isLetter(s.charAt(i)))
			{ result += s.charAt(i);
			}
		}
		return result;	
	}
	/*
	String trimPunctuation(String s)
	{	char c = s.charAt(s.length()-1);
		if(s.length()==0) return s;	
		while (c == '.' || c == ';' || c == ':' || c == ',')
		{	s = s.substring(0, s.length()-1);
			if(s.length()== 0) return s;
			c = s.charAt(s.length()-1);	
		}	
		return s;	
	}
*/

	// The real guts of the converter.
    String wikiToHTMLEntry(String s)throws IOException
    {
		// Set for one loop when a CR is encountered.
		boolean justCR = true;

		_in = new StringReader(s);
		_sBuf = new StringBuffer();

		while(true)
		{
			String frag = readFrag();
			if(frag == null)
				break; // Out of while loop.
			if(justCR && frag.indexOf("---") == 0)
			{	if(frag.indexOf("------") == 0)
				_sBuf.append("\n<HR size = \"10\">\n");
				else if(frag.indexOf("-----") == 0)
				_sBuf.append("\n<HR size = \"7\">\n");
				else if(frag.indexOf("----") == 0)
				_sBuf.append("\n<HR size = \"4\">\n");
				else
				_sBuf.append("\n<HR>\n");
			} 
			else if(frag.charAt(0) == CR && justCR)
			{
				_sBuf.append("\n<BR>\n");
			}
			else if( frag.indexOf("http:") == 0 ||
            		 frag.indexOf("shttp:") == 0 ||
					 frag.indexOf("mailto:") == 0 ||
					 frag.indexOf("news:") == 0 ||
					 frag.indexOf("ftp:") == 0 ||
					 frag.indexOf("mailto:") == 0 ||
					 frag.indexOf("gopher:") == 0 ||
					 frag.indexOf("telnet:") == 0)
			{
				_sBuf.append("\n<A HREF=\"" + frag + "\">" + frag + "</a>\n");
			}
			else if(isWikiLink(frag))
			{
				_sBuf.append("\n<A HREF=\"" + frag + "\">" + frag + "</a>\n");
			}
			else
			{
				_sBuf.append(frag);
			}


			if(frag.charAt(0) == CR)
			{
				justCR = true;
				_sBuf.append((char)CR);
			}
			else
			{
				justCR = false;
			}
				
		}

		return _sBuf.toString();
    }
	
	static String replaceDate(String s)
	{	String result = s;
		Date date = new Date();	
		int loc = s.indexOf("<date>");
		while (loc >= 0)
		{	result = result.substring(0, loc) +"(" + date + ")" + result.substring(loc+6);		
			loc = result.indexOf("<date>", loc);
		}
		return result;	
	}
    
	/** True if s has the first character upper case
	 ** then lower case, then later is upper case again.
	 ** They are composed entirely of letters and digits. 
	 ** TODO HtTp://... will not be handled correctly. 
	 ** This is a HREF in Wiki style docs.
	 **/
    boolean isWikiLink(String s)
    {
		if(s.length() < 3)
			return false;

		if( ! Character.isUpperCase(s.charAt(0)))
			return false;

		if(Character.isUpperCase(s.charAt(1)))
			return false;

		for(int i=2; i<s.length(); i++)
		{	char ch = s.charAt(i);
			// Here we avoid detecting HTML tags as case changes.
	/*		if( ch == '<' || 
				ch == '>' ||
				ch == ';' ||
				ch == '"' ||
				ch == '\''||
				ch == '&')
				return false;
	*/
			if(!Character.isLetterOrDigit(ch)) // JB 9-2003 (removed above also.)
				return false;
			if(Character.isUpperCase(ch))
				return true;
		}

		return false;
    } 
    
	/** Reads the next fragment of the input.
	 ** A frag is either a series of white space, a lone CR,
	 ** or a series of printables.
	 ** Things that might be wiki words are handled separately.
     ** Some special characters (" ') also terminate scanning. 
	 ** @returns null when no more input.
	 **/
	String readFrag()
		throws IOException
	{
		int c = getC();

		// We ignore LF's
		while (c == LF)
        {	c = getC();	
        }

		// End of String
		if(c == -1)
		{
			return null;
		}

		StringBuffer b = new StringBuffer();

		if(c == ' ' || c == '\t') // A chunk of whitespace.
		{
			while(c == ' ' || c == '\t') // A chunk of whitespace.
			{
				b.append((char)c);
				c = getC();
			}
			ungetC(c);
		}
        else if (c == '\'' || c == '"' ||  c == ',') // Joe Bergin 11/00
        {	b.append((char)c);
        }
		else if(c == CR) // End of line.
		{
			b.append((char)CR);
		}
		else // Read until a white space or special char.
		{	boolean mayBeWikiWord = Character.isUpperCase((char)c);// JB 9-2003
			boolean isWikiWord = false; // JB 9-2003
			
			while(c != ' ' 
            && c != '\t' 
            && c != CR 
            && c != LF 
            && c != '\''  // Joe Bergin 11/00
            && c != '"'   // Joe Bergin 11/00
           // && c != '.'   // Joe Bergin 11/00 // bug makes links not work. 
            && c != ','   // Joe Bergin 11/00
		
            && c != -1)
			{	if(!isWikiWord && !Character.isLetterOrDigit((char)c))
					mayBeWikiWord = false;
				if(b.length() == 1 && !Character.isLowerCase((char)c))
					mayBeWikiWord = false;
				else if(b.length()>1 && mayBeWikiWord && Character.isUpperCase((char)c))
					isWikiWord = true;
				else if(isWikiWord && !Character.isLetterOrDigit((char)c))
					break;
				b.append((char)c);
				c = getC();
				if(c == '<')break; // Joe Bergin 9/01 
			}
			ungetC(c);
		}
		return b.toString();
	}

	int getC()throws IOException
	{
		if(_ungetIsValid)
		{
			_ungetIsValid = false;
			return _ungetChar;
		}
		return _in.read();
	}

	void ungetC(int c)throws IOException
	{
		if(_ungetIsValid)
		{
			System.out.println("Warning: JWikiCvt unget overflow.");
		}
		_ungetIsValid = true;
		_ungetChar = c;
	}
    
	/** Reads up to 64k characters from the file specified by args[0]
	 ** then translates it and prints the results.  For debug use.
	 **/
    public static void main(String args[])
    {
	  try
	  {
		File f = new File(args[0]);
		FileInputStream in = new FileInputStream(f);

		byte[] buf = new byte[65535];
		int r = in.read(buf);
		in.close();

		String s = new String(buf, 0, r);

		s = JWikiCvt.wikiToHTML(s);

		System.out.println(s);
	  }
	  catch(Exception e)
	  {
         System.out.println("exception: " + e.getMessage());
         e.printStackTrace();	 
      }
    } 

}
