package jwiki;

import java.io.*;
import java.util.*;

/** Does the recent pages function of JWiki 
**	@author JoeBergin
**/
public class recent implements WikiPostable 
{
	private static final int maxHits = 100;

    public void handlePOST(JHandler h, byte buf[], int bufSize)
        throws Exception
    {		
        h.sendString("HTTP/1.0 200 Ok");
    	h.crlf();
    	// Headers
    	// This keeps the browser from expecting a persistant connection.
    	h.sendString("Connection: close");
    	h.crlf();      
    	h.crlf();
    	
    	h.startPage("Recently Changed Pages", false);
        
    	Vector fnames = getHits(h);
    	int size = fnames.size();
    	if(size == 0)
    	{	int recent = JWiki.recent();
    		h.sendString(("<H1>No Changes in "+ recent +" day" +(recent>1?"s":"")+".</H1>\n"));
    	}
    	else
    	{	qs(fnames, 0, size - 1);
    		// Show them but limit it to 100 hits.
            h.sendString(("" + size + " matching page" +(size>1?"s":"")+" found. Most recent changes first.<P>"));
    		h.sendString("<pre>");
        	for(int i=0; i < size && i < maxHits; i++)
    		{	TimePair file = (TimePair)fnames.elementAt(i);
    			String nm = file.name();
              	String pad = "";
                for(int j = nm.length(); j <30; ++j) pad += " ";
                String when = new Date(file.time()).toString();
    	    	h.sendString(("<A HREF=" + nm + ">" + nm +"</A> "+ pad+ when+"<BR>"));
    		}
            h.sendString("</pre>");
    	}

    	h.endPage();
        h.flush();
    }
    
    /** Get the names of files modified in the last recent days
    */
    private static Vector getHits(JHandler h)
    {	Vector result = new Vector();
    	File docs = new File(h.getParent().getRoot() + File.separator + "docs");
    	long baseTime = new Date().getTime() - 1000L*60L*60L*24L*(long)JWiki.recent();
         String fnames[] = docs.list();
    	for(int i=0; i<fnames.length; i++)
    	{	File file = new File(h.getParent().getRoot() + File.separator +"docs" +File.separator + (String)fnames[i]);
            long modTime = file.lastModified();
            if (baseTime <= modTime && file.isFile()) // No directories.
            {	result.addElement(new TimePair(modTime ,fnames[i]));
            }
        }
        
		return result;    
    }
    
    private static class TimePair implements Comparable
    {	private long timeInMillis;
    	private String fileName;
        
        long time()
        {	return timeInMillis;
        }
        
        String name()
        {	return fileName;
        }
        
        TimePair(long timeLastModified, String fileName)
        {	this.fileName = fileName;
            this.timeInMillis = timeLastModified;    
        }
        
        public int compareTo(Object other) // meaning is ">0 if this is earlier than other"
        {	if(timeInMillis < ((TimePair)other).timeInMillis) return 1;
    		if(timeInMillis > ((TimePair)other).timeInMillis) return -1;    
        	return 0;
        }       
    }
    
    /** Untuned recursive quicksort. Normally O(n*lg(n)), but can degenerate.
    ** @param v The vector to be sorted
    ** @param start The first cell in the portion to be sorted
    ** @param end The last cell in the portion to be sorted.  
    */
    static void qs(Vector v, int start, int end)
    {	if (start < end)
    	{	int m = partition(v, start, end);
    		qs(v, start, m-1);
            qs(v, m+1, end);    
    	}
    }
    
    /** Partition step for quicksort
    ** @param v The vector to be partitioned
    ** @param start The first cell in the portion to be partitioned
    ** @param end The last cell in the portion to be partitioned.  
    */
    private static int partition(Vector v, int start, int end)
    {	int m = start;
		Object temp;    
		Comparable o = (Comparable)v.elementAt(m);
        for (int t = start; t <= end; t++)
        {	if( ((Comparable)v.elementAt(t)).compareTo(o) < 0)
        	{	m++;
        		temp = v.elementAt(m);
        		v.setElementAt(v.elementAt(t), m);
        		v.setElementAt(temp, t);       		    
        	}           
        }
        temp = v.elementAt(m);
        v.setElementAt(v.elementAt(start), m);
        v.setElementAt(temp, start);
		return m;    
    }
    
/*    static void sort(Vector v, int start, int end)
    {
	    int i, j;
	    // Clip the heck out of it.
    	if((end + 1) <= start || 
       		start < 0 || end < 0 || 
       		start > (v.size() - 1) || end > v.size())
    		return;

    // Ouch, a bubble sort.  Some one needs to make this better.
    	for(i = end - 1; i >= 0; i--)
    	{
    		for(j=start; j < i; j++)
    		{
    			Comparable a = (Comparable)v.elementAt(j);
    			Comparable b = (Comparable)v.elementAt(j + 1);

    			if(a.compareTo(b) < 0)
    			{
    				// Swap
    				Object c = v.elementAt(j);
    				v.setElementAt(v.elementAt(j + 1), j);
    				v.setElementAt(c, j + 1);
    			}
    		}
    	}
    
    }
*/
}

