package jwiki;

import java.io.*;
import java.util.*;

/** Does the search function of JWiki **/
public class search implements WikiPostable {

	/** The most hits that are ever returned.  Limits searches
	 ** in both time and memory space.
	 **/
	private static int maxHits = 100;

    public void handlePOST(JHandler h, byte buf[], int bufSize)
        throws Exception
    {
        // Parse the content part of the request into seperate items.
        // The key is the var name and the data is the value of the
        // var.
        Hashtable vars = h.parsePOST(buf, bufSize);

        String whatfor = (String)vars.get("whatfor");
        if(whatfor == null || whatfor.length() == 0)
        {
            h.transmitString("Empty Search Error", "You must specify what to search for.");
            return;
        }

		// Search for the word in all the doc files.
		// Returns a Vector of all the file names.
		// The vector has the file names that match firs, then file contents that
		// match, in alphabetical order.
		Vector fnames = searchFor(h, whatfor);
       
        //OutputStream out = h.getOut();
        h.sendString("HTTP/1.0 200 Ok");
        h.crlf();
        // Headers
        // This keeps the browser from expecting a persistant connection.
        h.sendString("Connection: close");
        h.crlf();      
        h.crlf();
        
        h.startPage("Search Results",false);
 
 		if(fnames.size() == 0)
		{
        	h.sendString(("<H1>No Matches</H1>\n"));
		}
		else
		{
			// Show them but limit it to 100 hits.
	        h.sendString(("" + fnames.size() + " matching files found.<P>"));
			for(int i=0; i<fnames.size() && i < maxHits; i++)
			{
				String nm = (String)fnames.elementAt(i);
	        	h.sendString(("<A HREF=" + nm + ">" + nm + "</A><BR>"));
			}
		}

        h.endPage();
        h.flush();
    }

	/** Creates a Vector with all files names that match pat in the file
	 ** name or in contents.  Ignores case.  Not HTML savy.
	 ** The upper part of the Vector has the file name matches, and the lower part
	 ** has the file content matches.  Sorted in alphabetical order.
	 ** Never reurns null.  Can return an empty Vector.
	 **/
	public static Vector searchFor(JHandler h, String pat)
	{
		Vector v = new Vector(10, 10);
		int startOfCont;
		int i;

		pat = pat.toLowerCase();

		// First fill the vector with all file names in the doc directory that
		// have the word in them
		try
		{
			File docs = new File(h.getParent().getRoot() + File.separator + "docs");
			String fnames[] = docs.list();
			for(i=0; i<fnames.length; i++)
			{
				String low = fnames[i].toLowerCase();
				if(low.indexOf(pat) != -1) // A hit?
				{
					v.addElement(fnames[i]);
				}
			}
			startOfCont = v.size();
			qsStringVector(v, 0, startOfCont-1);

			// Now scan for content matches.
			for(i=0; i<fnames.length; i++)
			{
				if(matchContents(h, 
					new File(h.getParent().getRoot() + File.separator + "docs", fnames[i]), 
					pat))
				{
					v.addElement(fnames[i]);
				}
			}
			// Sort this section too.
			qsStringVector(v, startOfCont, v.size()-1);

			// Now remove duplicates.
			removeDups(v);
			
		}
		catch (Exception e)
		{
			v.addElement("Threw an exception while searching: " + e.getMessage());
		}

		return v;
	}

	/** Sorts the section of the Vector from start to end-1.
	 ** Assumes the Vector is all String.
	 ** start and end can be any number and will not cause an exception.
	 **/
/*	public static void sortVector(Vector v, int start, int end)
	{
		int i, j;

		// Clip the heck out of it.
		if((end + 1) <= start || 
		   start < 0 || end < 0 || 
		   start > (v.size() - 1) || end > v.size())
			return;

		// Ouch, a bubble sort.  Some one needs to make this better.
		for(i = end - 1; i >= 0; i--)
		{
			for(j=start; j < i; j++)
			{
				String a = (String)v.elementAt(j);
				String b = (String)v.elementAt(j + 1);

				a = a.toLowerCase();
				b = b.toLowerCase();
				if(a.compareTo(b) > 0)
				{
					// Swap
					Object c = v.elementAt(j);
					v.setElementAt(v.elementAt(j + 1), j);
					v.setElementAt(c, j + 1);
				}
			}
		}
	}
*/

	/** Untuned recursive quicksort. Normally O(n*lg(n)), but can degenerate.
    **  Sorts Strings. 
	** @param v The vector to be sorted
	** @param start The first cell in the portion to be sorted
	** @param end The last cell in the portion to be sorted.  
	*/
	static void qsStringVector(Vector v, int start, int end)
	{	if (start < end)
		{	int m = partition(v, start, end);
			qsStringVector(v, start, m-1);
	        qsStringVector(v, m+1, end);    
		}
	}
	
	/** Partition step for quicksort
	** @param v The vector to be partitioned
	** @param start The first cell in the portion to be partitioned
	** @param end The last cell in the portion to be partitioned.  
	*/
	static int partition(Vector v, int start, int end)
	{	int m = start;
		Object temp;    
		String o = ((String)v.elementAt(m)).toLowerCase();
	    for (int t = start; t <= end; t++)
	    {	if( ((String)v.elementAt(t)).toLowerCase().compareTo(o) < 0)
	    	{	m++;
	    		temp = v.elementAt(m);
	    		v.setElementAt(v.elementAt(t), m);
	    		v.setElementAt(temp, t);       		    
	    	}           
	    }
	    temp = v.elementAt(m);
	    v.setElementAt(v.elementAt(start), m);
	    v.setElementAt(temp, start);
		return m;    
	}
	

	/** Scans a file for matches of pat in it.  pat must be lower case.
	 ** @returns true if there is a match.
	 **/
	public static boolean matchContents(JHandler h, File f, String pat)
	{
		try
		{
			String contents = h.getFile(f);

			if(contents.toLowerCase().indexOf(pat) != -1)
			{
				return true;
			}

		}
		catch (Exception e)
		{
		}
		return false;
	}

	/** Removes any duplicate Strings in v.  v is assumed to be String. **/
	public static void removeDups(Vector v)
	{
		String a;
		String b;
		int i;
		int j;

		for(i=0; i<v.size(); i++)
		{
			a = (String)v.elementAt(i);

			for(j=i+1; j<v.size(); j++)
			{
				b = (String)v.elementAt(j);
				if(a.equals(b))
				{
					v.removeElementAt(j);
					j--;
				}
			}
		}
	}

}
